import builtins
from typing import Any, List, Optional

from requests import Response


class LivyError(Exception):
    '''
    Base class
    '''


class LivyStatementError(LivyError):
    '''
    Error occurred during the execution of a Livy statement
    '''
    def __init__(self, ename: str, evalue: str, traceback: List[str]):
        super().__init__(f'error in Livy statement execution: {ename}({evalue}):\n{"".join(traceback)}')
        self.ename = ename
        self.evalue = evalue
        self.traceback = traceback

    def as_builtin(self) -> Optional[BaseException]:
        '''
        Returns a built-in exception instance if the error type is a built-in
        '''
        exc_type = getattr(builtins, self.ename, None)
        if exc_type and issubclass(exc_type, BaseException):
            return exc_type(self.evalue)


class LivySessionDeadError(LivyError):
    '''
    Error occurred because the Livy session is dead
    '''
    def __init__(self, session_name: str):
        super().__init__(f'session {session_name!r} is no longer running')


class LivySessionBusyError(LivyError):
    '''
    Error occurred because the Livy session is dead
    '''
    def __init__(self, session_name: str):
        super().__init__(f'session {session_name!r} is no longer running')


class LivyRequestError(LivyError):
    '''
    Error occurred during the execution of a Livy request
    '''
    def __init__(self, response: Response, body: Any):
        message = f'unexpected response status={response.status_code}: {body!r}'
        super().__init__(message)

        self.response = response
        self.body = body


class LivyRetriableError(LivyError):
    '''
    Retriable error occurred during the execution of a Livy request
    '''
    pass



from typing import Optional
import pytest

from livy_uploads.utils import assert_type


@pytest.mark.parametrize(
    ['value', 'expected_type'],
    [
        (1, int),
        (1.0, float),
        ('foo', str),
        (True, bool),
        (None, Optional[int]),
        (None, Optional[float]),
        (None, Optional[str]),
    ],
)
def test_assert_type_good(value, expected_type):
    assert assert_type(value, expected_type) == value

@pytest.mark.parametrize(
    ['value', 'expected_type'],
    [
        (1, str),
        (1.0, str),
        (True, str),
        (None, str),
        (1, Optional[str]),
        (1.0, Optional[str]),
        (True, Optional[str]),
        (None, int),
    ],
)
def test_assert_type_bad(value, expected_type):
    with pytest.raises(ValueError):
        assert_type(value, expected_type)



from logging import getLogger
import threading
from typing import Any, Dict, Mapping, Optional, TypeVar

import requests
import requests.exceptions

from livy_uploads.retry_policy import RetryPolicy, DontRetryPolicy, WithExceptionsPolicy
from livy_uploads.exceptions import LivyRequestError, LivyRetriableError
from livy_uploads.utils import assert_type, try_decode
from livy_uploads.auth import Authenticator


LOGGER = getLogger(__name__)
T = TypeVar('T')


class LivyEndpoint:
    '''
    A class to upload generic data to a remote Spark session using the Livy API.
    '''
    def __init__(
        self,
        url: str,
        default_headers: Optional[Dict[str, str]] = None,
        verify: Optional[bool] = True,
        authenticator: Optional[Authenticator] = None,
        requests_session: Optional[requests.Session] = None,
        retry_policy: Optional[RetryPolicy] = None,
        proxy: Optional[str] = None,
    ):
        '''
        Parameters:
        - url: the base URL of the Livy server
        - default_headers: a dictionary of headers to include in every request
        - verify: whether to verify the SSL certificate of the server
        - authenticator: an optional authentication object factory to pass to requests
        - requests_session: an optional requests.Session object to use for making requests
        - retry_policy: an optional retry policy to use for requests
        '''
        self.url = url.rstrip('/')
    
        if default_headers is None:
            default_headers = {'content-type': 'application/json'}
        self.default_headers = {k.lower(): v for k, v in default_headers.items()}

        self.verify = True if verify is None else (verify or False)
        self._auth = None
        self.authenticator = authenticator
        self._auth_lock = threading.RLock()
        self.requests_session = requests_session or requests.Session()
        self.retry_policy = retry_policy or DontRetryPolicy()

        self.requests_session.trust_env = False
        self.proxy = proxy or None
        if self.proxy:
            self.requests_session.proxies.update({
                'http': self.proxy,
                'https': self.proxy,
            })

    @property
    def auth(self):
        if self.authenticator is None:
            return None

        if self._auth:
            return self._auth

        with self._auth_lock:
            if self._auth is None:
                self._auth = self.authenticator()
            return self._auth

    @classmethod
    def from_config(cls, config: Optional[Mapping[str, Any]]) -> 'LivyEndpoint':
        if not config:
            raise ValueError('config is required')

        session = requests.Session()
        session.trust_env = False
        proxy = assert_type(config.get('proxy'), Optional[str])
        if proxy:
            session.proxies.update({
                'http': proxy,
                'https': proxy,
            })

        return cls(
            url=assert_type(config['url'], str),
            default_headers=assert_type(config.get('default_headers'), Optional[dict]),
            verify=assert_type(config.get('verify'), Optional[bool]),
            authenticator=Authenticator.from_config(config.get('auth')),
            retry_policy=RetryPolicy.from_config(config.get('retry_policy')),
            proxy=proxy,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.url!r})'

    def build_headers(self, headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        '''
        Merges the list of default headers with the provided headers, and normalizes the keys to lowercase
        '''
        headers = {k.lower(): v for k, v in (headers or {}).items()}
        return {**self.default_headers, **headers}

    def request(
        self,
        method: str,
        path: str,
        headers=None,
        retry_policy: Optional[RetryPolicy] = None,
        **kwargs,
    ) -> requests.Response:
        '''
        Sends a request to the Livy endpoint.

        Parameters:
        - method: the HTTP method to use
        - path: the path to append to the base URL
        - headers: a dictionary of headers to include in the request. If None, the default headers will be used
        - retry_policy: an optional retry policy to use for this request, defaults to the one configured in the endpoint
        - kwargs: extra arguments to pass to `requests.Session.request`
        '''
        retry_policy = WithExceptionsPolicy(retry_policy or self.retry_policy, LivyRetriableError)
        if headers is None:
            headers = self.default_headers

        return self.retry_policy.run(
            func=_request_do,
            session=self.requests_session,
            method=method,
            url=self.url + path,
            headers=headers,
            auth=self.auth,
            verify=self.verify,
            **kwargs,
        )


def _request_do(
    session: requests.Session,
    *args,
    **kwargs,
) -> requests.Response:
    try:
        response = session.request(*args, **kwargs)
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
        raise LivyRetriableError from e

    if response.status_code < 300:
        return response

    if response.status_code == 429 or response.status_code >= 500:
        raise LivyRetriableError

    raise LivyRequestError(response, body=try_decode(response))



from abc import ABC, abstractmethod
import time
from typing import Any, Callable, Mapping, Optional, TypeVar

from livy_uploads.utils import assert_type

T = TypeVar('T')
E = TypeVar('E', bound=Exception)


class RetryPolicy(ABC):
    '''
    An abstract class for defining retry policies.
    '''

    def __new__(cls, *args, **kwargs):
        '''
        The __new__ method stores the arguments passed to the constructor so that the policy can be cloned later.
        '''
        instance = super().__new__(cls)
        instance._init_args = args
        instance._init_kwargs = kwargs
        return instance

    def clone(self) -> 'RetryPolicy':
        '''
        Clones this retry policy with its state reset.

        By default, this will recreate the policy with the same arguments passed to __init__.
        Reimplement this method if your class needs a custom reset behavior (mainly, if it receives a mutable argument)
        '''
        return self.__class__(*self._init_args, **self._init_kwargs)

    @abstractmethod
    def should_retry(self, exception: Exception) -> bool:
        '''
        Returns True if the exception should be retried.
        '''

    @abstractmethod
    def next_delay(self) -> float:
        '''
        Returns the delay before the next retry.
        '''

    def delay(self, seconds: float):
        '''
        Sleeps for the specified number of seconds.
        '''
        time.sleep(seconds)

    def run(self, func: Callable[..., T], *args, **kwargs) -> T:
        '''
        Runs the function with the specified retry policy.
        '''
        self = self.clone()

        while True:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if not self.should_retry(e):
                    raise
                self.delay(self.next_delay())

    @classmethod
    def from_config(cls, config: Optional[Mapping[str, Any]]) -> 'RetryPolicy':
        if not config:
            return DontRetryPolicy()
        return LinearRetryPolicy(
            max_tries=assert_type(config['max_tries'], int),
            pause=assert_type(config['pause'], float),
        )


class DontRetryPolicy(RetryPolicy):
    '''
    A retry policy that never retries.
    '''

    def should_retry(self, e: Exception) -> bool:
        return False

    def next_delay(self) -> float:
        return 0


class LinearRetryPolicy(RetryPolicy):
    '''
    A retry policy that retries a fixed number of times with a linear delay.
    '''

    def __init__(self, max_tries: int, pause: float):
        self.max_tries = max_tries
        self.pause = pause
        self._current_try = 1

    def should_retry(self, e: Exception) -> bool:
        return self._current_try < self.max_tries

    def next_delay(self) -> float:
        self._current_try += 1
        return self.pause


class TimeoutRetryPolicy(RetryPolicy):
    '''
    A retry policy that retries until a timeout is reached.
    '''
    def __init__(self, timeout: float, pause: float):
        self.timeout = timeout
        self.pause = pause
        self._start_time = time.monotonic()

    def should_retry(self, e: Exception) -> bool:
        return time.monotonic() - self._start_time < self.timeout

    def next_delay(self) -> float:
        return self.pause


class WithExceptionsPolicy(RetryPolicy):
    '''
    A derived retry policy that only retries on a specific set of exceptions.
    '''

    def __init__(self, base: RetryPolicy, *exceptions: E):
        if isinstance(base, WithExceptionsPolicy):
            base = base.base

        self.base = base
        self.exceptions = exceptions

    def clone(self) -> 'WithExceptionsPolicy':
        return self.__class__(self.base.clone(), *self.exceptions)

    def should_retry(self, e: Exception) -> bool:
        if not isinstance(e, self.exceptions):
            return False
        return self.base.should_retry(e)

    def next_delay(self) -> float:
        return self.base.next_delay()



from functools import wraps
import json
from pathlib import Path
import re
import time
import traceback
from typing import Iterator, List, Optional

from IPython.core.getipython import get_ipython
from IPython.core.interactiveshell import InteractiveShell
from sparkmagic.livyclientlib.sparkcontroller import SparkController
from sparkmagic.livyclientlib.livyreliablehttpclient import LivyReliableHttpClient
from IPython.core.magic import magics_class, Magics
from IPython.core.magic import needs_local_scope, line_magic, cell_magic
from IPython.core.magic_arguments import argument, magic_arguments
from IPython.core.magic_arguments import parse_argstring
from sparkmagic.utils.sparklogger import SparkLog
from sparkmagic.livyclientlib.exceptions import (
    handle_expected_exceptions,
    wrap_unexpected_exceptions,
    BadUserDataException,
    SparkStatementException,
    LivyClientLibException,
)
from hdijupyterutils.ipythondisplay import IpythonDisplay

from livy_uploads.session import LivySession
from livy_uploads.endpoint import LivyEndpoint
from livy_uploads.commands import (
    LivyRunCode,
    LivyRunShell,
    LivyUploadDir,
    LivyUploadFile,
)


def wrap_standard_exceptions(f):
    @wraps(f)
    def inner(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except LivyClientLibException as e:
            raise
        except Exception as e:
            traceback.print_exc()
            raise SparkStatementException('error') from e
    return inner


@magics_class
class LivyUploaderMagics(Magics):
    def __init__(self, shell):
        super().__init__(shell)
        self.ipython_display = IpythonDisplay()
        self.logger = SparkLog("LivyUploaderMagics")

    @magic_arguments()
    @argument(
        "-n",
        "--varname",
        type=str,
        default=None,
        help="Name of the variable to send.",
    )
    @argument(
        "-s",
        "--session-name",
        type=str,
        default=None,
        help="Name of the Livy client to use. If not provided, uses the default one",
    )
    @line_magic
    @needs_local_scope
    @wrap_unexpected_exceptions
    @handle_expected_exceptions
    @wrap_standard_exceptions
    def send_obj_to_spark(self, line, cell="", local_ns=None):
        if cell.strip():
            raise BadUserDataException(
                "Cell body for %%{} magic must be empty; got '{}' instead".format(
                    'send_obj_to_spark', cell.strip()
                )
            )
        args = parse_argstring(LivyUploaderMagics.send_obj_to_spark, line)
        if not args.varname:
            raise BadUserDataException(
                "Variable name must be provided with -n/--varname option"
            )
        session = get_session(getattr(args, 'session_name', None))
        cmd = LivyRunCode(
            vars=dict(
                varname=args.varname,
                value=(local_ns or {})[args.varname],
            ),
            code='''
                globals()[varname] = value
            '''
        )
        session.apply(cmd)

    @magic_arguments()
    @argument(
        "-n",
        "--varname",
        type=str,
        default=None,
        help="Name of the variable to fetch.",
    )
    @argument(
        "-s",
        "--session-name",
        type=str,
        default=None,
        help="Name of the Livy client to use. If not provided, uses the default one",
    )
    @line_magic
    @needs_local_scope
    @wrap_unexpected_exceptions
    @handle_expected_exceptions
    def get_obj_from_spark(self, line, cell="", local_ns=None):
        if cell.strip():
            raise BadUserDataException(
                "Cell body for %%{} magic must be empty; got '{}' instead".format(
                    'send_obj_to_spark', cell.strip()
                )
            )
        args = parse_argstring(LivyUploaderMagics.get_obj_from_spark, line)
        if not args.varname:
            raise BadUserDataException(
                "Variable name must be provided with -n/--varname option"
            )
        session = get_session(getattr(args, 'session_name', None))
        cmd = LivyRunCode(
            vars=dict(varname=args.varname),
            code='''
                _ = globals()[varname]
            '''
        )
        _, result = session.apply(cmd)
        local_ns[args.varname] = result

    @magic_arguments()
    @argument(
        "-p",
        "--path",
        type=str,
        default=None,
        help="Local path to upload.",
    )
    @argument(
        "-d",
        "--dest",
        type=str,
        default=None,
        help="Remote path to upload into.",
    )
    @argument(
        "-c",
        "--chunk-size",
        type=int,
        default=50_000,
        help="Max size in each upload chunk",
    )
    @argument(
        "-m",
        "--mode",
        type=int,
        default=0,
        help="Permissions to set on the uploaded file or directory. Defaults to 0o700 for directories and 0o600 for files.",
    )
    @argument(
        "-s",
        "--session-name",
        type=str,
        default=None,
        help="Name of the Livy client to use. If not provided, uses the default one",
    )
    @line_magic
    @needs_local_scope
    @wrap_unexpected_exceptions
    @handle_expected_exceptions
    @wrap_standard_exceptions
    def send_path_to_spark(self, line, cell="", local_ns=None):
        if cell.strip():
            raise BadUserDataException(
                "Cell body for %%{} magic must be empty; got '{}' instead".format(
                    'send_obj_to_spark', cell.strip()
                )
            )
        args = parse_argstring(LivyUploaderMagics.send_path_to_spark, line)
        if not args.path:
            raise BadUserDataException(
                "Source must be provided with -s/--source option"
            )
        source = Path(args.path)
        if not source.exists():
            raise BadUserDataException(
                f"Source path {source} does not exist"
            )

        session = get_session(getattr(args, 'session_name', None))
        if source.is_dir():
            cmd = LivyUploadDir(
                source_path=args.path,
                dest_path=args.dest,
                chunk_size=args.chunk_size,
                mode=args.mode or 0o700,
            )
        else:
            cmd = LivyUploadFile(
                source_path=args.path,
                dest_path=args.dest,
                chunk_size=args.chunk_size,
                mode=args.mode or 0o600,
            )

        final_path = session.apply(cmd)
        self.ipython_display.write(f"Uploaded {source} to {final_path}")

    @magic_arguments()
    @argument(
        "-s",
        "--session-name",
        type=str,
        default=None,
        help="Name of the Livy client to use. If not provided, uses the default one",
    )
    @argument(
        "-t",
        "--timeout",
        type=float,
        default=10.0,
        help="Max execution time of the command",
    )
    @needs_local_scope
    @cell_magic
    @wrap_unexpected_exceptions
    @handle_expected_exceptions
    def shell_command(self, line, cell="", local_ns=None):
        args = parse_argstring(LivyUploaderMagics.shell_command, line)
        command = cell.strip()

        if not command:
            raise BadUserDataException(
                "Non-empty command must be provided in the cell"
            )

        session = get_session(getattr(args, 'session_name', None))
        cmd = LivyRunShell(
            command=command,
            run_timeout=args.timeout,
        )
        output, returncode = session.apply(cmd)
        for l in output.splitlines():
            print(l)
        self.ipython_display.write(f"$ command exited with code {returncode}")
        local_ns = local_ns or {}
        local_ns['shell_output'] = output
        local_ns['shell_returncode'] = returncode

    _logs_follower: Iterator[List[str]] = None

    @magic_arguments()
    @argument(
        "-s",
        "--session-name",
        type=str,
        default=None,
        help="Name of the Livy client to use. If not provided, uses the default one",
    )
    @argument(
        "-p",
        "--page-size",
        type=int,
        default=100,
        help="Max lines to fetch at once",
    )
    @argument(
        "-r",
        "--reset",
        action="store_true",
        default=False,
        help="Reset the logs follower",
    )
    @needs_local_scope
    @cell_magic
    @wrap_unexpected_exceptions
    @handle_expected_exceptions
    def logs_follow(self, line, cell="", local_ns=None):
        args = parse_argstring(LivyUploaderMagics.logs_follow, line)
        command = cell.strip()

        if command:
            raise BadUserDataException(
                "No command must be provided in the cell"
            )

        if self.__class__._logs_follower is None or args.reset:
            session = get_session(getattr(args, 'session_name', None))
            self.__class__._logs_follower = session.follow(
                page_size=args.page_size,
            )

        try:
            lines = next(self.__class__._logs_follower)
        except StopIteration:
            lines = []

        local_ns['logs_lines'] = lines

        if not lines:
            self.ipython_display.write(f"No new logs")
            return

        for line in lines:
            print(line)

def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(LivyUploaderMagics)


def get_session(session_name: Optional[str] = None) -> 'LivySession':
    '''
    Creates a session endpoint instance from the current IPython shell
    '''
    cell_magics = get_ipython().magics_manager.magics['cell']

    try:
        magic_name = 'send_to_spark'
        spark_magic = cell_magics['send_to_spark']
    except KeyError:
        try:
            magic_name = 'spark'
            spark_magic = cell_magics['spark']
        except KeyError:
            raise RuntimeError("no spark magic found named '%send_to_spark' or '%spark'")

    try:
        spark_controller: SparkController = spark_magic.__self__.spark_controller
    except (AttributeError, TypeError):
        raise RuntimeError(f'no spark controller found in magic %{magic_name!r}')

    livy_session = spark_controller.get_session_by_name_or_default(session_name)
    livy_client = livy_session.http_client._http_client

    return LivySession(
        url=livy_client._endpoint.url,
        session_id=livy_session.id,
        default_headers=livy_client._headers,
        verify=livy_client.verify_ssl,
        requests_session=livy_client._session,
    )



from base64 import b64decode, b64encode
from io import BytesIO
from logging import getLogger
import os
import pickle
import shutil
import textwrap
from tempfile import TemporaryDirectory
import time
from typing import Any, Callable, Dict, List, Optional, TypeVar, Tuple
from uuid import uuid4

from livy_uploads.exceptions import LivyStatementError
from livy_uploads.session import LivySession, LivyCommand


LOGGER = getLogger(__name__)
T = TypeVar('T')


class LivyRunCode(LivyCommand[Tuple[List[str], Any]]):
    '''
    Executes the code in the global namespace of the remote Livy session.

    Assigns the return value from the `_` variable.
    '''

    def __init__(self, code: str, pause: float = 0.3, vars: Optional[Dict[str, Any]] = None):
        self.code = code
        self.pause = pause
        self.vars = vars or {}

    def run(self, session: 'LivySession') -> Any:
        var_lines = [
            'import pickle',
            'import base64',
            '_ = None',
        ]

        for var_name, var_value in self.vars.items():
            pickled_b64 = b64encode(pickle.dumps(var_value)).decode('ascii')
            var_lines.append(f'{var_name} = pickle.loads(base64.b64decode({repr(pickled_b64)}))')

        return_lines = [
            "pickled_b64 = base64.b64encode(pickle.dumps(_)).decode('ascii')",
            "print('\\nLivyUploads:pickled_b64', len(pickled_b64), pickled_b64, end='\\n')",
        ]

        code = '\n'.join(var_lines) + '\n' + textwrap.dedent(self.code) + '\n' + '\n'.join(return_lines)

        r = session.request(
            'POST',
            f"/sessions/{session.session_id}/statements",
            json={
                'kind': 'pyspark',
                'code': code,
            },
        )
        st_id = r.json()['id']

        st_path = f"/sessions/{session.session_id}/statements/{st_id}"
        headers = session.build_headers({'accept': 'application/json'})

        while True:
            r = session.request('GET', st_path, headers=headers)
            st = r.json()
            if st['state'] in ('waiting', 'running'):
                time.sleep(self.pause)
                continue
            elif st['state'] == 'available':
                break
            else:
                raise Exception(f'statement failed to execute: {st}')

        output = st['output']
        if output['status'] != 'ok':
            raise LivyStatementError(
                output['ename'],
                output['evalue'],
                output['traceback'],
            )
        try:
            lines: List[str] = output['data']['text/plain'].strip().splitlines()
        except KeyError:
            raise Exception(f'non-textual output: {output}')

        try:
            final_lines: List[str] = []
            for line in lines:
                if not line.startswith('LivyUploads:pickled_b64'):
                    final_lines.append(line)
                    continue

                parts = line.strip().split()
                try:
                    prefix, size, data_b64 = parts
                    size = int(size)
                except Exception as e:
                    raise RuntimeError(f'bad status line in {line!r}') from e

                if size != len(data_b64):
                    raise ValueError(
                        f'bad output, len does not match (expected {len(data_b64)}, got {size}: {lines}'
                    )
                if prefix != 'LivyUploads:pickled_b64':
                    raise ValueError(f'bad output, unexpected prefix {prefix!r}')

                value = pickle.loads(b64decode(data_b64))
                return lines[:-1], value

        except Exception as e:
            raise Exception(f'bad output, unexpected format: {output}') from e


class LivyUploadBlob(LivyCommand):
    '''
    Uploads a file-like object to the remote Spark session.

    The path to file can then be obtained remotely with `pyspark.SparkFiles.get(name)`.
    '''

    def __init__(self, source: BytesIO, name: str):
        '''
        Parameters:
        - source: the file-like object to read from
        - name: the name of the uploaded file
        '''
        self.source = source
        self.name = name

    def run(self, session: 'LivySession'):
        '''
        Executes the upload
        '''
        headers = session.build_headers({'accept': 'application/json'})
        headers.pop('content-type', None)

        session.request(
            'POST',
            f'/sessions/{session.session_id}/upload-file',
            headers=headers,
            files={'file': (self.name, self.source)},
        )


class LivyUploadFile(LivyCommand[str]):
    '''
    Uploads a (potentially large) file in chunks to a path in the remote Spark session.
    '''

    def __init__(self, source_path: str, dest_path: Optional[str] = None, chunk_size: int = 50_000, mode: int = 0o600, progress_func: Optional[Callable[[float], None]] = None):
        '''
        Parameters:
        - source_path: the path to the file to upload
        - dest_path: the path where the file will be saved in the remote session. If not provided, the basename of the source path will be used.
        - chunk_size: the size of the chunks to split the file into
        - mode: the permissions to set on the file after reassembly
        - progress_func: an optional function that will be called with a float between 0 and 1 to indicate the progress of the upload
        '''
        self.source_path = source_path
        self.dest_path = dest_path or os.path.basename(os.path.abspath(source_path))
        self.chunk_size = chunk_size
        self.mode = mode
        self.progress_func = progress_func or (lambda _: None)

    def run(self, session: 'LivySession') -> str:
        '''
        Executes the uploading of the file to the remote Spark session.

        The file will be split into chunks of the specified size, uploaded and reassembled in the remote session.
        '''
        file_size = os.stat(self.source_path).st_size
        num_chunks = (file_size + self.chunk_size - 1) // self.chunk_size
        basename = f'upload-file-{uuid4()}'

        with open(self.source_path, 'rb') as source:
            i = 0
            while True:
                chunk = source.read(self.chunk_size)
                if not chunk:
                    break
                upload_cmd = LivyUploadBlob(BytesIO(chunk), f'{basename}.{i}')
                upload_cmd.run(session)
                i += 1
                self.progress_func(i / num_chunks)

        num_chunks = i

        merge_cmd = LivyRunCode(
            vars=dict(
                basename=basename,
                num_chunks=num_chunks,
                dest_path=self.dest_path,
                mode=self.mode,
            ),
            code=f'''
                import os
                import os.path
                import pyspark

                os.makedirs(os.path.dirname(dest_path) or '.', exist_ok=True)
                with open(dest_path, 'wb') as fp:
                    pass
                os.chmod(dest_path, mode)

                with open(dest_path, 'wb') as fp:
                    for i in range(num_chunks):
                        chunk_name = f'{{basename}}.{{i}}'
                        with open(pyspark.SparkFiles.get(chunk_name), 'rb') as chunk_fp:
                            fp.write(chunk_fp.read())

                _ = os.path.realpath(dest_path)
            ''',
        )

        _, path = merge_cmd.run(session)
        return path


class LivyUploadDir(LivyCommand[str]):
    '''
    Uploads a (potentially large) directory in chunks to a path in the remote Spark session.
    '''

    def __init__(self, source_path: str, dest_path: Optional[str] = None, chunk_size: int = 50_000, mode: int = 0o700, progress_func: Optional[Callable[[float], None]] = None):
        '''
        Parameters:
        - source_path: the path to the directory to upload
        - dest_path: the path where the directory will be saved in the remote session. If not provided, the basename of the source path will be used.
        - chunk_size: the size of the chunks to split the archive into
        - mode: the permissions to set on the directory after extraction
        - progress_func: an optional function that will be called with a float between 0 and 1 to indicate the progress of the upload
        '''
        if not source_path:
            raise ValueError('source_path must be provided')
        self.source_path = source_path
        self.dest_path = dest_path or os.path.basename(os.path.abspath(source_path))
        self.chunk_size = chunk_size
        self.mode = mode or 0o700
        self.progress_func = progress_func or (lambda _: None)

    def run(self, session: 'LivySession') -> str:
        '''
        Executes the uploading of the directory to the remote Spark session.

        The directory will be archived, split into chunks of the specified size, uploaded, reassembled and extracted in the remote session.
        '''
        archive_name = f'archive-{uuid4()}'

        with TemporaryDirectory() as tempdir:
            archive_source = shutil.make_archive(
                base_name=os.path.join(tempdir, archive_name),
                format='gztar',
                root_dir=self.source_path,
            )
            archive_dest = f'tmp/{os.path.basename(archive_source)}'
            upload_cmd = LivyUploadFile(
                source_path=archive_source,
                dest_path=archive_dest,
                chunk_size=self.chunk_size,
                progress_func=self.progress_func,
            )
            upload_cmd.run(session)

        extract_cmd = LivyRunCode(
            vars=dict(
                archive_dest=archive_dest,
                dest_path=self.dest_path,
                mode=self.mode,
            ),
            code=f'''
                import os
                import shutil

                try:
                    try:
                        shutil.rmtree(dest_path)
                    except FileNotFoundError:
                        pass

                    shutil.unpack_archive(archive_dest, dest_path)
                    os.chmod(dest_path, mode)
                finally:
                    try:
                        os.remove(archive_dest)
                    except FileNotFoundError:
                        pass

                _ = os.path.realpath(dest_path)
            ''',
        )

        _, path = extract_cmd.run(session)
        return path


class LivyRunShell(LivyCommand[Tuple[str, int]]):
    '''
    Executes a shell command in the remote Spark session.
    '''

    def __init__(self, command: str, run_timeout: float = 5.0, stop_timeout: float = 2.0):
        '''
        Parameters:
        - command: the shell command to execute
        - run_timeout: the maximum time to wait for the command to complete. SIGTERM will be
        sent to process if this time is exceeded
        - stop_timeout: time to wait after sending SIGTERM to wait for the process to properly die
        '''
        self.command = command
        self.run_timeout = run_timeout
        self.stop_timeout = stop_timeout

    def run(self, session: 'LivySession') -> Tuple[str, int]:
        '''
        Executes the command and returns the output and the return code.
        '''
        code_cmd = LivyRunCode(
            vars=dict(
                command=self.command,
                run_timeout=self.run_timeout,
                stop_timeout=self.stop_timeout,
            ),
            code='''
                import subprocess

                proc = subprocess.Popen(
                    ['bash', '-c', command],
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    universal_newlines=True,
                )
                try:
                    proc.wait(run_timeout)
                except subprocess.TimeoutExpired:
                    proc.terminate()
                    try:
                        proc.wait(stop_timeout)
                    except suprocess.TimeoutExpired:
                        proc.kill()

                _ = proc.stdout.read(), proc.poll()
            ''',
        )
        _, (output, returncode) = code_cmd.run(session)
        return output, returncode



from typing import Any, Type, TypeVar, Union

import requests

T = TypeVar('T')


def try_decode(response: requests.Response) -> Any:
    '''
    Tries to decode the response as JSON or text.
    '''
    try:
        return response.json()
    except requests.exceptions.JSONDecodeError:
        try:
            return response.text
        except UnicodeDecodeError:
            return response.content.decode('utf8', errors='replace')


def assert_type(value: Any, expected_type: Type[T]) -> T:
    """
    Type assertion utility function.
    """
    try:
        origin = getattr(expected_type, '__origin__')
        if origin is Union:
            args = expected_type.__args__
            if len(args) == 2 and args[1] is type(None):
                nullable = True
                expected_type = args[0]
    except AttributeError:
        nullable = False

    if nullable and value is None:
        return value

    if not isinstance(value, expected_type):
        raise ValueError(f'Expected {expected_type}, got {type(value)}')

    return value



import importlib.util
import logging
from pathlib import Path
import shutil
from typing import List, Optional, Union
import zipfile


LOGGER = logging.getLogger(__name__)


def pack_package(package: str, output_dir: Union[str, Path], globs: Optional[List[str]] = None) -> Path:
    '''
    Packs a Python package into a ZIP file.

    Args:
        package: The top-level package name to be packed. Must not contain dots.
        output_dir: The path to the output directory.
        globs: The glob patterns to use to find the package files. Defaults to ['*.py'].

    Returns:
        Path to the created ZIP file
    '''
    if globs is None:
        globs = ['*.py']

    if '.' in package:
        raise ValueError('package must be a single module')

    output_dir = Path(output_dir).absolute()
    if output_dir.is_file():
        raise ValueError('output_dir must be a directory')

    output_path = output_dir / f'{package}.zip'
    spec = importlib.util.find_spec(package)
    if not spec:
        raise FileNotFoundError(f'package {package} not found')

    origin = Path(spec.origin or '')
    if not spec.origin or not origin.exists():
        raise IOError(f'package {package} has no file origin')

    tmp_path = output_path.parent / f'.{output_path.name}.tmp'

    if origin.name.endswith('.zip'):
        LOGGER.info('using ZIP package at %s', origin)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(origin, tmp_path)
    else:
        with zipfile.ZipFile(tmp_path, 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
            if origin.name == '__init__.py':
                LOGGER.info('recursively adding package directory at %s', origin.parent)
                for glob_pattern in globs:
                    for path in origin.parent.rglob(glob_pattern):
                        if path.is_file():
                            zipf.write(path, f'{package}/{path.relative_to(origin.parent)}')
            else:
                LOGGER.info('adding single package file at %s', origin)
                zipf.write(origin, f'{package}.py')

    shutil.move(tmp_path, output_path)
    LOGGER.info('ZIP package saved to %s', output_path)
    return output_path




import pytest
from unittest.mock import Mock, patch

from livy_uploads.retry_policy import DontRetryPolicy, LinearRetryPolicy, WithExceptionsPolicy


class TestWithExceptionsLinearRetryPolicy:
    base = LinearRetryPolicy(3, 0.01)
    policy = WithExceptionsPolicy(base, ValueError)

    def test_wrapping_behavior(self):
        wrapped = WithExceptionsPolicy(self.policy, IndexError)
        assert wrapped.base is self.policy.base

    def test_no_errors(self):
        f = Mock()
        self.policy.run(f)
        assert f.call_count == 1

    @patch('time.sleep')
    def test_unrecognized_error(self, patched_time_sleep: Mock):
        f = Mock()
        f.side_effect = [ValueError(), IndexError(), 42]

        with pytest.raises(IndexError):
            self.policy.run(f)

        assert f.call_count == 2
        patched_time_sleep.assert_called_once_with(0.01)

    @patch('time.sleep')
    def test_recognized_error(self, patched_time_sleep: Mock):
        f = Mock()
        f.side_effect = [ValueError(), ValueError(), 42]

        assert self.policy.run(f) == 42
        assert f.call_count == 3
        assert patched_time_sleep.call_count == 2

    @patch('time.sleep')
    def test_exhaustion(self, patched_time_sleep: Mock):
        f = Mock()
        f.side_effect = [ValueError()] * 3

        with pytest.raises(ValueError):
            self.policy.run(f)

        assert f.call_count == 3
        assert patched_time_sleep.call_count == 2



class TestLinearRetryPolicy:
    policy = LinearRetryPolicy(2, 0.01)

    def test_no_errors(self):
        f = Mock()
        self.policy.run(f)
        assert f.call_count == 1

    @patch('time.sleep')
    def test_one_error(self, patched_time_sleep: Mock):
        f = Mock()
        f.side_effect = [IndexError(), 42]

        assert self.policy.run(f) == 42
        assert f.call_count == 2
        patched_time_sleep.assert_called_once_with(0.01)

    @patch('time.sleep')
    def test_exhaustion(self, patched_time_sleep: Mock):
        f = Mock()
        f.side_effect = [ValueError()] * 2

        with pytest.raises(ValueError):
            self.policy.run(f)

        assert f.call_count == 2
        assert patched_time_sleep.call_count == 1


class TestDontRetryPolicy:
    def test_no_errors(self):
        retry_policy = DontRetryPolicy()
        f = Mock()
        retry_policy.run(f)
        assert f.call_count == 1

    @patch('time.sleep')
    def test_error(self, patched_time_sleep: Mock):
        retry_policy = DontRetryPolicy()
        f = Mock()
        f.side_effect = [Exception(), 42]

        with pytest.raises(Exception):
            retry_policy.run(f)

        assert f.call_count == 1
        assert patched_time_sleep.call_count == 0



import logging
from threading import Event, Thread
import textwrap
import time
from uuid import uuid4

import pytest

from livy_uploads.exceptions import LivyRequestError
from livy_uploads.session import LivyEndpoint, LivySession
from livy_uploads.retry_policy import LinearRetryPolicy


logging.basicConfig(
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s:%(funcName)s: %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TestLivySessionEndpoint:
    endpoint = LivyEndpoint('http://localhost:8998')

    def test_lifecycle(self):
        sessions_before = {s.session_id: s for s in LivySession.list(self.endpoint)}

        name = 'test-' + str(uuid4())
        session = LivySession.create(self.endpoint, name=name, ttl='60s', heartbeatTimeoutInSecond=60)

        sessions_now = {s.session_id: s for s in LivySession.list(self.endpoint)}
        new_session_ids = list(set(sessions_now) - set(sessions_before))
        assert len(new_session_ids) == 1
        new_session = sessions_now[new_session_ids[0]]

        assert new_session.session_name == name
        assert new_session.session_id == session.session_id

        # WARNING: Livy seems to fail to cleanup if we close the session too quickly
        session.wait_ready(LinearRetryPolicy(30, 1.0))

        with pytest.raises(LivyRequestError) as e:
            LivySession.create(self.endpoint, name=name)

        assert 'duplicate' in str(e).lower()

        session.delete()
        session.wait_done(LinearRetryPolicy(10, 1.0))
        assert session.refresh_state() == 'dead'

    def test_ttl(self):
        session = LivySession.create(
            self.endpoint,
            name=f'test-{uuid4()}',
            ttl='10s',
            heartbeatTimeoutInSecond=10,
        )
        try:

            stopped = Event()
            do_refresh = True
            exc = None

            def run():
                try:
                    while do_refresh:
                        session.refresh_state()
                        time.sleep(1.0)
                except Exception as e:
                    nonlocal exc
                    exc = e
                finally:
                    stopped.set()

            Thread(target=run, daemon=True).start()

            session.wait_ready(LinearRetryPolicy(30, 1.0))

            do_refresh = False
            if not stopped.wait(10.0):
                raise RuntimeError('session refresh thread did not stop as expected')

            logging.info('waiting 1m30s to check session expires')
            time.sleep(90)

            assert session.refresh_state() == 'dead'
        finally:
            try:
                session.delete()
            except Exception as e:
                logging.exception('error deleting session: %s', e)

    def test_follow(self):
        session = LivySession.create(
            self.endpoint,
            name=f'test-{uuid4()}',
            ttl='60s',
            heartbeatTimeoutInSecond=60,
        )
        session.wait_ready(LinearRetryPolicy(60, 1.0))
        try:

            logs_iter = session.follow()
            creation_logs = []
            for logs in logs_iter:
                if not logs:
                    break
                creation_logs += logs
            assert 'Created Spark session.' in '\n'.join(creation_logs)

            code = textwrap.dedent('''
                from datetime import datetime
                import time
                import traceback

                from pyspark import InheritableThread

                def run():
                    println = sc._gateway.jvm.java.lang.System.err.println
                    try:
                        while True:
                            now = datetime.now().astimezone().isoformat()
                            println('test_follow: ' + now)
                            time.sleep(1.0)
                    except Exception:
                        println(traceback.format_exc())

                thread = InheritableThread(target=run, daemon=True)
                thread.start()
            ''')
            r = session.request(
                'POST',
                f"/sessions/{session.session_id}/statements",
                json={
                    'kind': 'pyspark',
                    'code': code,
                },
            )

            seen = set(creation_logs)
            time.sleep(5.0)
            logs = next(logs_iter)
            assert len(logs) < 30
            assert all(log not in seen for log in logs if log.strip() not in ["stderr:", "stdout:"])

            seen.update(logs)
            time.sleep(5.0)
            logs = next(logs_iter)
            assert len(logs) < 30
            assert all(log not in seen for log in logs if log.strip() not in ["stderr:", "stdout:"])
        finally:
            try:
                session.delete()
            except Exception as e:
                logging.exception('error deleting session: %s', e)



from abc import ABC, abstractmethod
from datetime import datetime, timezone
import itertools
from logging import getLogger, Logger
from threading import RLock
import time
import uuid
from typing import Any, Dict, Generic, Iterator, List, Mapping, Optional, Set, TypeVar

import requests

from livy_uploads.exceptions import LivyError, LivyRequestError, LivyRetriableError
from livy_uploads.endpoint import LivyEndpoint
from livy_uploads.retry_policy import RetryPolicy, WithExceptionsPolicy
from livy_uploads.utils import assert_type
from livy_uploads.auth import Authenticator


LOGGER = getLogger(__name__)
T = TypeVar('T')


class LivySession(LivyEndpoint):
    '''
    A class to query and interact with a specific Livy session.
    '''

    def __init__(
        self,
        url: str,
        session_id: int,
        session_info: Optional[Dict[str, Any]] = None,
        default_headers: Optional[Dict[str, str]] = None,
        verify: bool = True,
        authenticator: Optional[Authenticator] = None,
        requests_session: Optional[requests.Session] = None,
        retry_policy: Optional[RetryPolicy] = None,
        logger: Optional[Logger] = None,
    ):
        '''
        Parameters:
        - url: the base URL of the Livy server
        - session_id: the ID of the Spark session to use
        - default_headers: a dictionary of headers to include in every request
        - verify: whether to verify the SSL certificate of the server
        - authenticator: an optional authentication object to pass to requests
        - requests_session: an optional requests.Session object to use for making requests
        - retries: the number of times to retry a request if it fails
        - pause: the number of seconds to wait between polling for the status of a statement
        '''
        super().__init__(
            url=url,
            default_headers=default_headers,
            verify=verify,
            authenticator=authenticator,
            requests_session=requests_session,
            retry_policy=retry_policy,
        )
        self.session_id = session_id
        self.session_info = session_info or {}
        self.session_info['id'] = session_id
        self.lock = RLock()
        self.logger = logger or getLogger(f'session#{session_id}')

    def __repr__(self):
        app_info = self.session_info.get('appInfo') or {}
        infos = ', '.join([
            f'url={self.url}',
            f'id={self.session_id}',
            f'name={self.session_name}',
            f'state={self.session_state}',
            f'app_id={self.session_info.get("appId") or None}',
            f'ui_url={app_info.get("sparkUiUrl")}',
            f'driver_url={app_info.get("driverLogUrl")}',
        ])
        return f'{self.__class__.__name__}({infos})'

    @classmethod
    def list(self, endpoint: LivyEndpoint, page_size: int = 20) -> Iterator['LivySession']:
        '''
        Retrieves a list of all active session endpoints
        '''
        offset = 0

        while True:
            LOGGER.info('fetching sessions from offset %d', offset)
            r = endpoint.request('GET', f'/sessions?from={offset}&size={page_size}')
            body = dict(r.json())
            sessions = body.get('sessions') or []
            offset += len(sessions)

            for session in sessions:
                yield LivySession(
                    url=endpoint.url,
                    session_id=session['id'],
                    session_info=session,
                    default_headers=endpoint.default_headers,
                    verify=endpoint.verify,
                    authenticator=endpoint.authenticator,
                    requests_session=endpoint.requests_session,
                )
            if not sessions or len(sessions) < page_size:
                break

    @classmethod
    def from_config(cls, config: Optional[Mapping[str, Any]]) -> 'LivySession':
        if not config:
            raise ValueError('config is required')
        
        endpoint = LivyEndpoint.from_config(config.get('endpoint'))
        kwargs = assert_type(config['session'], dict)
        name = assert_type(kwargs.pop('name'), str).format(
            timestamp=datetime.now(timezone.utc).isoformat(),
            uuid=str(uuid.uuid4()),
        )

        for session in cls.list(endpoint):
            if session.session_name == name:
                LOGGER.info('reusing existing session: %s', session)
                return session

        LOGGER.info('no existing session found, creating new one')
        return cls.create(endpoint=endpoint, name=name, **kwargs)

    @classmethod
    def create(cls, endpoint: 'LivyEndpoint', name: str, doAs: Optional[str] = None, **kwargs) -> 'LivySession':
        '''
        Creates a new session in the Livy server

        Parameters:
        - endpoint: the Livy endpoint to use
        - name: the name of the session
        - doAs: an optional user to impersonate
        - kwargs: additional parameters to post to Livy. Check `POST /sessions` in the Livy API docs
        (https://livy.incubator.apache.org/docs/latest/rest-api.html#session) for more information.
        '''
        path = '/sessions'
        if doAs:
            path += f'?doAs={doAs}'

        LOGGER.info('creating new session named %r', name)
        r = endpoint.request(
            'POST',
            path='/sessions',
            json={**kwargs, 'name': name},
        )
        body = r.json()

        session = LivySession(
            url=endpoint.url,
            session_id=body['id'],
            default_headers=endpoint.default_headers,
            verify=endpoint.verify,
            authenticator=endpoint.authenticator,
            requests_session=endpoint.requests_session,
            retry_policy=endpoint.retry_policy,
        )

        time.sleep(1.0)  # just in case, in tests it seems to be kind of flaky
        session.refresh()
        session.logger.info('session created')
        return session

    @property
    def session_state(self) -> str:
        '''
        Returns the state of the session
        '''
        return self.session_info.get('state') or '<unknown>'

    @property
    def session_name(self) -> str:
        '''
        Returns the name of the session
        '''
        return self.session_info.get('name') or '<unknown>'

    def assert_idle(self):
        '''
        Asserts that the session is in the idle state
        '''
        self.refresh_state()

        if self.session_state != 'idle':
            raise LivyError(f'session is not idle: {self.session_state!r}')

    def wait_ready(self, retry_policy: RetryPolicy):
        '''
        Waits until the session is ready
        '''
        retry_policy = WithExceptionsPolicy(retry_policy, LivyError)

        def run():
            state = self.refresh_state()
            if state == 'idle':
                self.logger.info('session is ready')
                return
            if state in ('not_started', 'starting', 'busy'):
                self.logger.info('session is still in state %r, waiting...', state)
                raise LivyError
            if state != 'idle':
                raise RuntimeError(f'bad session state: {state!r}')

        retry_policy.run(run)

    def delete(self):
        '''
        Sends the deletion request for the session
        '''
        self.logger.info('stopping session')
        self.request('DELETE', f'/sessions/{self.session_id}')
        self.logger.info('session deleted')

    def wait_done(self, retry_policy: RetryPolicy):
        '''
        Waits until the session is finished
        '''
        retry_policy = WithExceptionsPolicy(retry_policy, LivyRetriableError)

        def run():
            state = self.refresh_state()
            if state == 'dead':
                self.logger.info('session is deleted')
                return
            elif state in ('killed', 'success', 'error'):
                self.logger.info('session finished in state %r')
                return

            self.logger.info('session is still in state %r, waiting...', state)
            raise LivyRetriableError

        retry_policy.run(run)

    def refresh(self) -> Dict[str, Any]:
        '''
        Refreshes all the information for the session
        '''
        r = self.request('GET', f'/sessions/{self.session_id}')
        body = dict(r.json())
        body.pop('log', None)
        self.session_info = body
        return self.session_info

    def refresh_state(self) -> str:
        '''
        Refreshes the state of the session
        '''

        try:
            r = self.request('GET', f'/sessions/{self.session_id}/state')
            body = dict(r.json())
            state = body['state']
        except LivyRequestError as e:
            if e.response.status_code == 404:
                state = 'dead'
            else:
                raise

        self.session_info['state'] = state
        return state

    def follow(self, page_size: int = 500) -> Iterator[List[str]]:
        '''
        Iterates over the logs of the session.

        Livy rotates the logs returned in the API, so this method uses a very crude heuristic to try to
        detect the shift and deduplicate them.

        Parameters:
        - page_size: the number of log lines to read at once

        Yields:
        - the batch of log lines. If it returns one with zero lines, you probably should add a bigger pause
        before the next iteration to avoid hitting the server too hard.
        '''
        previous_logs_set: Set[str] = set()

        while True:
            state = self.refresh_state()
            if state in ('error', 'dead', 'killed', 'success'):
                LOGGER.info('session is in state %r, stopping log retrieval', state)
                break

            try:
                r = self.request('GET', f'/sessions/{self.session_id}/log?size={page_size}')
                body = dict(r.json())
            except LivyRequestError as e:
                if e.response.status_code == 404:
                    LOGGER.info('session %d is not found, stopping log retrieval', self.session_id)
                    break
                else:
                    raise

            current_logs: List[str] = body.get('log') or []
            current_logs_set: Set[str] = set(current_logs)
            logs: List[str] = list(itertools.dropwhile(lambda l: l in previous_logs_set, current_logs))
            previous_logs_set = current_logs_set

            yield logs

    def apply(self, command: 'LivyCommand[T]') -> T:
        '''
        Runs a command in the remote Livy session.
        '''
        with self.lock:
            self.assert_idle()
            return command.run(self)


class LivyCommand(ABC, Generic[T]):
    '''
    Base class for Livy commands.
    '''
    @abstractmethod
    def run(self, session: LivySession) -> T:
        '''
        Runs the command in the specified Livy session.
        '''



from datetime import datetime, timedelta
import hashlib
import os
from pathlib import Path
import shlex
import signal
import time
from uuid import uuid4

import pytest
from unittest.mock import Mock


from livy_uploads.exceptions import LivyStatementError
from livy_uploads.session import LivyEndpoint, LivySession
from livy_uploads.retry_policy import LinearRetryPolicy
from livy_uploads.commands import (
    LivyRunCode,
    LivyUploadFile,
    LivyUploadDir,
    LivyRunShell,
)

class TestCommands:
    endpoint = LivyEndpoint('http://localhost:8998')
    session: LivySession

    @classmethod
    def setup_class(cls):
        cls.session = LivySession.create(
            cls.endpoint,
            name='test-' + str(uuid4()),
            ttl='60s',
            heartbeatTimeoutInSecond=60,
        )
        cls.session.wait_ready(LinearRetryPolicy(30, 1.0))

    def test_run(self):
        now = datetime.now().astimezone()

        code_cmd = LivyRunCode(
            vars=dict(
                now=now,
            ),
            code='''
                from datetime import timedelta
                _ = now + timedelta(days=1)
            ''',
        )
        lines, out = code_cmd.run(self.session)
        assert not lines
        assert isinstance(out, datetime)

        actual: datetime = out
        assert actual == now + timedelta(days=1)

    def test_run_exception(self):
        code_cmd = LivyRunCode(
            code='''
                int('invalid')
            ''',
        )

        with pytest.raises(LivyStatementError) as e:
            code_cmd.run(self.session)

        assert isinstance(e.value.as_builtin(), ValueError)

    def test_run_return_unpickleable(self):
        code_cmd = LivyRunCode(
            code='''
                import socket
                _ = socket.socket()
            ''',
        )

        with pytest.raises(LivyStatementError) as e:
            code_cmd.run(self.session)

        assert isinstance(e.value.as_builtin(), TypeError)

    def test_upload_file(self, tmp_path: Path):
        data = os.urandom(4096)
        path = tmp_path / 'data.bin'
        path.write_bytes(data)

        mock = Mock()
        dest_path = f'tmp/data-{uuid4()}.bin'
        upload_cmd = LivyUploadFile(
            source_path=path,
            dest_path=dest_path,
            chunk_size=len(data) // 4,
            progress_func=mock,
        )
        actual_path = upload_cmd.run(self.session)
        assert mock.call_count == 4

        test_cmd = LivyRunCode(
            vars=dict(
                dest_path=dest_path,
            ),
            code='''
                import hashlib
                import os
                import stat

                _ = (
                    os.getcwd(),
                    hashlib.md5(open(dest_path, 'rb').read()).hexdigest(),
                    oct(stat.S_IMODE(os.stat(dest_path).st_mode)),
                )
            ''',
        )
        _, (chdir, actual_md5, mode) = test_cmd.run(self.session)

        expected_md5 = hashlib.md5(data).hexdigest()

        assert actual_path == chdir + '/' + dest_path
        assert actual_md5 == expected_md5
        assert mode == '0o600'

    def test_upload_dir(self, tmp_path: Path):
        (tmp_path / 'foo').write_text('foo')
        (tmp_path / 'inner').mkdir()
        (tmp_path / 'inner' / 'bar').write_text('bar')

        dest_path = f'tmp/dir-{uuid4()}'
        upload_cmd = LivyUploadDir(
            source_path=tmp_path,
            dest_path=dest_path,
        )
        upload_cmd.run(self.session)

        test_cmd = LivyRunShell(f'find {shlex.quote(dest_path)} -type f')
        output, returncode = test_cmd.run(self.session)
        lines = list(sorted(output.splitlines()))

        assert returncode == 0
        assert lines[0].endswith('/foo')
        assert lines[1].endswith('/inner/bar')
        assert len(lines) == 2

    def test_shell_timeout(self):
        test_cmd = LivyRunShell(f'sleep 10', run_timeout=3, stop_timeout=2)
        t0 = time.monotonic()
        output, returncode = test_cmd.run(self.session)
        dt = time.monotonic() - t0
        lines = list(sorted(output.splitlines()))

        assert lines == []
        assert returncode == -int(signal.SIGTERM)
        assert 3 <= dt < 10



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from enum import Enum
import logging
import os
from pathlib import Path
from typing import Any, Mapping, Optional, Union
import subprocess

from requests.auth import AuthBase
from livy_uploads.utils import assert_type
try:
    from requests_gssapi import HTTPSPNEGOAuth
except ImportError:
    HTTPSPNEGOAuth = None


LOGGER = logging.getLogger(__name__)



class Authenticator:
    def __call__(self) -> Optional[AuthBase]:
        return None

    @classmethod
    def from_config(cls, config: Optional[Mapping[str, Any]]) -> 'Authenticator':
        if not config or config.get('type') != 'kerberos':
            return Authenticator()

        return KerberosAuthenticator(
            principal=assert_type(config.get('principal'), Optional[str]),
            keytab=assert_type(config.get('keytab'), Optional[str]),
            password=assert_type(config.get('password'), Optional[str]),
            krb5_config=assert_type(config.get('krb5_config'), Optional[str]),
            krb5_cache=assert_type(config.get('krb5_cache'), Optional[str]),
            mutual_authentication=getattr(MutualAuth, (assert_type(config.get('mutual_authentication'), Optional[str]) or 'OPTIONAL').upper()),
            target_name=assert_type(config.get('target_name'), Optional[str]),
            delegate=assert_type(config.get('delegate'), Optional[bool]),
            opportunistic_auth=assert_type(config.get('opportunistic_auth'), Optional[bool]),
        )


class MutualAuth(Enum):
    REQUIRED = 1
    OPTIONAL = 2
    DISABLED = 3


class KerberosAuthenticator(Authenticator):
    def __init__(
        self,
        principal: Optional[str] = None,
        keytab: Optional[Union[str, Path]] = None,
        password: Optional[str] = None,
        krb5_config: Optional[Union[str, Path]] = None,
        krb5_cache: Optional[Union[str, Path]] = None,
        mutual_authentication: Optional[MutualAuth] = None,
        target_name: Optional[str] = None,
        delegate: bool = False,
        opportunistic_auth: bool = False,
    ):
        if HTTPSPNEGOAuth is None:
            raise ImportError('requests-gssapi is not installed')

        self.principal = principal or None

        if self.principal:
            if not keytab and not password:
                raise ValueError('Either keytab or password must be provided')
            if keytab and password:
                raise ValueError('Only one of keytab or password must be provided')

        self.keytab = keytab or None
        self.password = password or None
        self.krb5_config = Path(krb5_config or '/etc/krb5.conf')
        self.krb5_cache = Path(krb5_cache or 'var/krb5_cache')
        self.mutual_authentication = mutual_authentication or None
        self.target_name = target_name or None
        self.delegate = delegate or None
        self.opportunistic_auth = opportunistic_auth or None

    def __call__(self) -> AuthBase:
        assert HTTPSPNEGOAuth is not None, 'requests-gssapi is not installed'

        LOGGER.info('initializing environment')
        self.krb5_cache.absolute().parent.mkdir(parents=True, exist_ok=True)
        os.environ['KRB5_CONFIG'] = str(self.krb5_config.absolute())
        os.environ['KRB5CCNAME'] = str(self.krb5_cache.absolute())

        if not self.principal:
            LOGGER.info('using cached credentials')
        elif self.keytab:
            LOGGER.info('logging with principal=%s and keytab=%s', self.principal, self.keytab)
            subprocess.run(['kinit', '-Vkt', self.keytab, self.principal], check=True)
        else:
            LOGGER.info('logging with principal=%s and password', self.principal)
            stdin = (self.password + '\n').encode('utf8')
            subprocess.run(['kinit', '-V', self.principal], check=True, input=stdin)

        kwargs = dict(
            mutual_authentication=self.mutual_authentication.value,
            target_name=self.target_name,
            delegate=self.delegate,
            opportunistic_auth=self.opportunistic_auth,
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return HTTPSPNEGOAuth(**kwargs)



from pathlib import Path

import papermill
import pytest


@pytest.mark.skip(reason="needs to configure sparkmagic")
def test_example_magics(request):
    rootdir = request.config.rootdir

    input_path = Path(rootdir) / "examples" / "magics.ipynb"
    output_path = Path(rootdir) / "tmp" / "output.ipynb"

    output_path.parent.mkdir(parents=True, exist_ok=True)

    papermill.execute_notebook(
        input_path=str(input_path),
        output_path=str(output_path),
        cwd=str(input_path.parent),
        progress_bar=False,
    )
